@@ -8,7 +8,6 @@ import traceback
 from contextlib import contextmanager
 from datetime import datetime
 from enum import Enum
-from functools import wraps
 from itertools import islice
 from queue import Queue
 from time import sleep, time
@@ -19,13 +18,15 @@ import grpc
 import health_pb2_grpc
 import numpy as np
 import requests
+from client_common.grpc_retry import grpc_retry
 
 import spm_pb2
 import spm_pb2_grpc
 
 from . import config
 from .common import JobStatus, RepeatedFunction, RestorableIterator, initialize_grpc_channel, log, parse_api_url
-from .data_source import DynamicSource, InputManager, StaticSource
+from .data_source import DataSourceCompleted, DataSourceEmptyDataset, DataSourceHibernating, DynamicSource, \
+    InputManager, StaticSource
 from .numerous_buffered_writer import NumerousBufferedWriter
 from .numerous_log_handler import NumerousLogHandler
 
@@ -34,25 +35,7 @@ try:
 except ModuleNotFoundError:
     pass
 
-
-def retry(func):
-    @wraps(func)
-    def wrapper(self, *args, **kwargs):
-        for attempt_number in range(1, config.GRPC_RETRY_ATTEMPTS + 1):
-            try:
-                return func(self, *args, **kwargs)
-            except grpc.RpcError as e:
-                if e.code() in (grpc.StatusCode.UNAVAILABLE, grpc.StatusCode.INTERNAL):
-                    if attempt_number == config.GRPC_RETRY_ATTEMPTS:
-                        log.warning("Exceeded command retry attempts")
-                        raise e
-                    timeout = config.GRPC_RETRY_DELAY * 2 ** attempt_number
-                    log.warning(f"Got error: {e}, wait for {timeout} seconds and retry the command")
-                    sleep(timeout)
-                else:
-                    raise e
-
-    return wrapper
+retry = grpc_retry(max_attempts=config.GRPC_RETRY_ATTEMPTS, start_delay=config.GRPC_RETRY_DELAY)
 
 
 def to_indexed_data_list(data_gen: Iterable[spm_pb2.DataList]) -> Iterator[spm_pb2.IndexedDataList]:
@@ -110,7 +93,7 @@ class JobParameters:
         scenario_data = spm_stub.GetScenario(spm_pb2.Scenario(project=project_id, scenario=scenario_id))
         job_info = json.loads(scenario_data.scenario_document)["jobs"][job_id]
         self._params: Dict[str, Any] = {}
-        for param in job_info["image"]["parameters"]:
+        for param in job_info["image"].get("parameters", {}):
             self._params[param["id"]] = param["value"]
 
     def get(self, key: str, default: Any):
@@ -127,6 +110,70 @@ class RunSettings:
         self.duration = self.end - self.start
 
 
+class DataSourcesReader:
+    def __init__(self,
+                 client,
+                 scenario_data: Dict,
+                 start_time: float,
+                 t0: float,
+                 te: float,
+                 dt: float,
+                 tag_prefix: str,
+                 tag_separator: str,
+                 timeout: float):
+        self._client = client
+        self._inputs_params = {'scenario_data': scenario_data, 't0': t0, 'te': te, 'dt': dt,
+                               'tag_prefix': tag_prefix, 'tag_separator': tag_separator, 'timeout': timeout}
+        self._reader: InputManager = self._client.get_inputs(**self._inputs_params)
+        self._t0 = t0
+        self._start_time = start_time
+        self._retry_delay: float = 1.0
+        self._retry_timeout: float = 10.0
+
+    def read(self, t: float):
+        empty_since = None
+        data = None
+        while data is None:
+            try:
+                if self._client.hibernate_event.is_set() or self._client.terminate_event.is_set():
+                    log.debug("Hibernate event or terminate event is set")
+                    return None
+
+                data = self._reader.get_at_time(t)
+                if data is None:
+                    sleep(self._retry_delay)
+                    continue
+                return data
+            except DataSourceHibernating:
+                log.debug("Data source hibernating, t0 = %f, start_time = %f", self._t0, self._start_time)
+                if self._t0 > self._start_time:
+                    self._reader = self._client.get_inputs(**self._inputs_params)
+                    sleep(self._retry_delay)
+                    continue
+                else:
+                    self._client.hibernate()
+                    return None
+            except DataSourceCompleted:
+                log.debug("Data source completed")
+                return None
+            except DataSourceEmptyDataset:
+                if empty_since is None:
+                    empty_since = time()
+                empty_duration = time() - empty_since
+                if empty_duration > self._retry_timeout:
+                    log.warning("Empty dataset for %s. Stopping", empty_duration)
+                    return None
+                self._reader = self._client.get_inputs(**self._inputs_params)
+                sleep(self._retry_delay)
+        return None
+
+    def set_retry_delay(self, value: float):
+        self._retry_delay = value
+
+    def set_retry_timeout(self, value: float):
+        self._retry_timeout = value
+
+
 class NumerousClient:
 
     def __init__(self, job_id: Optional[str] = None,
@@ -135,7 +182,8 @@ class NumerousClient:
                  refresh_token: Optional[str] = None,
                  clear_data: Optional[bool] = None, no_log: bool = False,
                  instance_id: Optional[str] = None,
-                 secure: Optional[bool] = None, url: Optional[str] = None):
+                 secure: Optional[bool] = None, url: Optional[str] = None,
+                 log_level: Union[str, int] = logging.ERROR):
         if url:
             server, port, secure = parse_api_url(url)
         self._t = [0]
@@ -199,7 +247,7 @@ class NumerousClient:
 
         self._terminated = False
         if not no_log:
-            self._init_logger()
+            self._init_logger(log_level)
 
         self.last_progress: float = -999999
         self.progress_debounce = 10
@@ -274,15 +322,16 @@ class NumerousClient:
     def _get_current_token(self):
         return self._access_token
 
-    def _init_logger(self):
+    def _init_logger(self, log_level):
         log_queue = Queue(-1)
         numerous_logger = logging.handlers.QueueHandler(queue=log_queue)
         logger = logging.getLogger()
-        logger.setLevel(logging.ERROR)
+        logger.setLevel(log_level)
         logger.addHandler(numerous_logger)
         self.log_handler = NumerousLogHandler(self)
         self.log_listener = logging.handlers.QueueListener(log_queue, self.log_handler)
         self.log_listener.start()
+        self.logger = logger
 
     @retry
     def _refresh_access_token(self, refresh_token, instance_id, execution_id, project_id=None, scenario_id=None,
@@ -627,7 +676,7 @@ class NumerousClient:
 
         last_row = 0
         last_time = time()
-        while time() - last_time < config.GRPC_RETRY_TIMEOUT:
+        while True:
             row = last_row
             try:
                 data_stream = self._base_client.ReadData(
@@ -644,7 +693,10 @@ class NumerousClient:
                     yield data
                 break
 
-            except grpc.RpcError:
+            except grpc.RpcError as e:
+                if time() - last_time >= config.GRPC_RETRY_TIMEOUT:
+                    log.warning(f"Exceeded read data stream retry timeout: {config.GRPC_RETRY_TIMEOUT}")
+                    raise e
                 last_row = row
                 sleep(config.GRPC_RETRY_DELAY)
 
@@ -766,6 +818,22 @@ class NumerousClient:
 
         log.debug('Reading as row complete')
 
+    def get_data_sources_reader(
+            self,
+            t0: float,
+            te: float,
+            scenario_data: Optional[Dict] = None,
+            dt: float = 3600.0,
+            tag_prefix: str = 'Stub',
+            tag_separator: str = '.',
+            timeout: float = 10.0,
+            start_time: Optional[float] = None) -> DataSourcesReader:
+        if scenario_data is None:
+            scenario_data, _ = self.get_scenario_document()
+        if start_time is None:
+            start_time = self.run_settings.start.timestamp()
+        return DataSourcesReader(self, scenario_data, start_time, t0, te, dt, tag_prefix, tag_separator, timeout)
+
     def clear_timeseries_data(self, scenario=None, execution=None):
         log.warning('Clearing data!')
 
@@ -928,7 +996,7 @@ class NumerousClient:
             writer.close()
 
     def get_inputs(self, scenario_data: dict, t0: float = 0, te: float = 0, dt: float = 3600, tag_prefix='Stub',
-                   tag_seperator: str = '.', timeout: int = 10):
+                   tag_separator: str = '.', timeout: int = 10):
         input_sources = {}
         static_data = {}
         input_source_types = {}
@@ -947,7 +1015,7 @@ class NumerousClient:
             if component["disabled"] or not component["isMainComponent"] and not component["isSubcomponent"]:
                 continue
             for inputVariable in component["inputVariables"]:
-                tag = tag_seperator.join(filter(None, [tag_prefix, component["name"], inputVariable["id"]]))
+                tag = tag_separator.join(filter(None, [tag_prefix, component["name"], inputVariable["id"]]))
                 datasource_type = inputVariable["dataSourceType"]
                 if datasource_type == "static":
                     static_data[tag] = inputVariable["value"]
@@ -1044,7 +1112,6 @@ class NumerousClient:
 
         # If job has been hibernated, call the set hibernate callback
         if self._hibernating:
-            self.state.commit()
             self._hibernate_callback(*self._hibernate_callback_args[0], **self._hibernate_callback_args[1])
 
         if self._complete:
@@ -1054,7 +1121,6 @@ class NumerousClient:
             self.complete_execution()
 
         if hasattr(self, 'log_handler'):
-            log.debug('flushing log')
             self.log_handler.close()
 
         log.debug('Client closing')
@@ -1064,6 +1130,7 @@ class NumerousClient:
         for cancel_func in self._cancel_streams:
             cancel_func()
 
+        self.state.commit()
         log.debug('Closing channel')
         self.channel.close()
 
